@@ -15,6 +15,7 @@
 #include "soup.h"
 #include "soup-connection.h"
 #include "soup-message-private.h"
+#include "soup-message-metrics-private.h"
 #include "soup-uri-utils-private.h"
 
 /**
@@ -85,13 +86,16 @@ typedef struct {
 	GUri              *first_party;
 	GUri              *site_for_cookies;
 
-	GTlsCertificate      *tls_certificate;
-	GTlsCertificateFlags  tls_certificate_errors;
+	GTlsCertificate      *tls_peer_certificate;
+	GTlsCertificateFlags  tls_peer_certificate_errors;
 
 	SoupMessagePriority priority;
 
 	gboolean is_top_level_navigation;
         gboolean is_options_ping;
+        guint    last_connection_id;
+
+        SoupMessageMetrics *metrics;
 } SoupMessagePrivate;
 
 G_DEFINE_TYPE_WITH_PRIVATE (SoupMessage, soup_message, G_TYPE_OBJECT)
@@ -132,8 +136,8 @@ enum {
 	PROP_FIRST_PARTY,
 	PROP_REQUEST_HEADERS,
 	PROP_RESPONSE_HEADERS,
-	PROP_TLS_CERTIFICATE,
-	PROP_TLS_CERTIFICATE_ERRORS,
+	PROP_TLS_PEER_CERTIFICATE,
+	PROP_TLS_PEER_CERTIFICATE_ERRORS,
 	PROP_PRIORITY,
 	PROP_SITE_FOR_COOKIES,
 	PROP_IS_TOP_LEVEL_NAVIGATION,
@@ -167,13 +171,14 @@ soup_message_finalize (GObject *object)
 	g_clear_pointer (&priv->uri, g_uri_unref);
 	g_clear_pointer (&priv->first_party, g_uri_unref);
 	g_clear_pointer (&priv->site_for_cookies, g_uri_unref);
+        g_clear_pointer (&priv->metrics, soup_message_metrics_free);
 
 	g_clear_object (&priv->auth);
 	g_clear_object (&priv->proxy_auth);
 
 	g_clear_pointer (&priv->disabled_features, g_hash_table_destroy);
 
-	g_clear_object (&priv->tls_certificate);
+	g_clear_object (&priv->tls_peer_certificate);
 
 	soup_message_headers_unref (priv->request_headers);
 	soup_message_headers_unref (priv->response_headers);
@@ -265,11 +270,11 @@ soup_message_get_property (GObject *object, guint prop_id,
 	case PROP_RESPONSE_HEADERS:
 		g_value_set_boxed (value, priv->response_headers);
 		break;
-	case PROP_TLS_CERTIFICATE:
-		g_value_set_object (value, priv->tls_certificate);
+	case PROP_TLS_PEER_CERTIFICATE:
+		g_value_set_object (value, priv->tls_peer_certificate);
 		break;
-	case PROP_TLS_CERTIFICATE_ERRORS:
-		g_value_set_flags (value, priv->tls_certificate_errors);
+	case PROP_TLS_PEER_CERTIFICATE_ERRORS:
+		g_value_set_flags (value, priv->tls_peer_certificate_errors);
 		break;
 	case PROP_PRIORITY:
 		g_value_set_enum (value, priv->priority);
@@ -559,8 +564,8 @@ soup_message_class_init (SoupMessageClass *message_class)
 	/**
 	 * SoupMessage::accept-certificate:
 	 * @msg: the message
-	 * @tls_certificate: the peer's #GTlsCertificate
-	 * @tls_errors: the tls errors of @tls_certificate
+	 * @tls_peer_certificate: the peer's #GTlsCertificate
+	 * @tls_peer_errors: the tls errors of @tls_certificate
 	 *
 	 * Emitted during the @msg's connection TLS handshake
 	 * after an unacceptable TLS certificate has been received.
@@ -708,30 +713,30 @@ soup_message_class_init (SoupMessageClass *message_class)
 				    G_PARAM_READABLE |
 				    G_PARAM_STATIC_STRINGS));
 	/**
-	 * SoupMessage:tls-certificate:
+	 * SoupMessage:tls-peer-certificate:
 	 *
-	 * The #GTlsCertificate associated with the message
+	 * The peer's #GTlsCertificate associated with the message
 	 *
 	 */
 	g_object_class_install_property (
-		object_class, PROP_TLS_CERTIFICATE,
-		g_param_spec_object ("tls-certificate",
-				     "TLS Certificate",
-				     "The TLS certificate associated with the message",
+		object_class, PROP_TLS_PEER_CERTIFICATE,
+		g_param_spec_object ("tls-peer-certificate",
+				     "TLS Peer Certificate",
+				     "The TLS peer certificate associated with the message",
 				     G_TYPE_TLS_CERTIFICATE,
 				     G_PARAM_READABLE |
 				     G_PARAM_STATIC_STRINGS));
 	/**
-	 * SoupMessage:tls-certificate-errors:
+	 * SoupMessage:tls-peer-certificate-errors:
 	 *
-	 * The verification errors on #SoupMessage:tls-certificate
+	 * The verification errors on #SoupMessage:tls-peer-certificate
 	 *
 	 */
 	g_object_class_install_property (
-		object_class, PROP_TLS_CERTIFICATE_ERRORS,
-		g_param_spec_flags ("tls-certificate-errors",
-				    "TLS Certificate Errors",
-				    "The verification errors on the message's TLS certificate",
+		object_class, PROP_TLS_PEER_CERTIFICATE_ERRORS,
+		g_param_spec_flags ("tls-peer-certificate-errors",
+				    "TLS Peer Certificate Errors",
+				    "The verification errors on the message's TLS peer certificate",
 				    G_TYPE_TLS_CERTIFICATE_FLAGS, 0,
 				    G_PARAM_READABLE |
 				    G_PARAM_STATIC_STRINGS));
@@ -1288,20 +1293,20 @@ soup_message_get_uri_for_auth (SoupMessage *msg)
 }
 
 static void
-soup_message_set_tls_certificate (SoupMessage         *msg,
-                                  GTlsCertificate     *tls_certificate,
-                                  GTlsCertificateFlags tls_errors)
+soup_message_set_tls_peer_certificate (SoupMessage         *msg,
+                                       GTlsCertificate     *tls_certificate,
+                                       GTlsCertificateFlags tls_errors)
 {
         SoupMessagePrivate *priv = soup_message_get_instance_private (msg);
 
-        if (priv->tls_certificate == tls_certificate && priv->tls_certificate_errors == tls_errors)
+        if (priv->tls_peer_certificate == tls_certificate && priv->tls_peer_certificate_errors == tls_errors)
                 return;
 
-        g_clear_object (&priv->tls_certificate);
-        priv->tls_certificate = tls_certificate ? g_object_ref (tls_certificate) : NULL;
-        priv->tls_certificate_errors = tls_errors;
-        g_object_notify (G_OBJECT (msg), "tls-certificate");
-        g_object_notify (G_OBJECT (msg), "tls-certificate-errors");
+        g_clear_object (&priv->tls_peer_certificate);
+        priv->tls_peer_certificate = tls_certificate ? g_object_ref (tls_certificate) : NULL;
+        priv->tls_peer_certificate_errors = tls_errors;
+        g_object_notify (G_OBJECT (msg), "tls-peer-certificate");
+        g_object_notify (G_OBJECT (msg), "tls-peer-certificate-errors");
 }
 
 SoupConnection *
@@ -1312,11 +1317,43 @@ soup_message_get_connection (SoupMessage *msg)
 	return priv->connection;
 }
 
+static void
+soup_message_set_metrics_timestamp_for_network_event (SoupMessage       *msg,
+                                                      GSocketClientEvent event)
+{
+        switch (event) {
+        case G_SOCKET_CLIENT_RESOLVING:
+                soup_message_set_metrics_timestamp (msg, SOUP_MESSAGE_METRICS_DNS_START);
+                break;
+        case G_SOCKET_CLIENT_RESOLVED:
+                soup_message_set_metrics_timestamp (msg, SOUP_MESSAGE_METRICS_DNS_END);
+                break;
+        case G_SOCKET_CLIENT_CONNECTING:
+                soup_message_set_metrics_timestamp (msg, SOUP_MESSAGE_METRICS_CONNECT_START);
+                break;
+        case G_SOCKET_CLIENT_CONNECTED:
+                /* connect_end happens after proxy and tls */
+        case G_SOCKET_CLIENT_PROXY_NEGOTIATING:
+        case G_SOCKET_CLIENT_PROXY_NEGOTIATED:
+                break;
+        case G_SOCKET_CLIENT_TLS_HANDSHAKING:
+                soup_message_set_metrics_timestamp (msg, SOUP_MESSAGE_METRICS_TLS_START);
+                break;
+        case G_SOCKET_CLIENT_TLS_HANDSHAKED:
+                break;
+        case G_SOCKET_CLIENT_COMPLETE:
+                soup_message_set_metrics_timestamp (msg, SOUP_MESSAGE_METRICS_CONNECT_END);
+                break;
+        }
+}
+
 static void
 re_emit_connection_event (SoupMessage       *msg,
                           GSocketClientEvent event,
                           GIOStream         *connection)
 {
+        soup_message_set_metrics_timestamp_for_network_event (msg, event);
+
 	g_signal_emit (msg, signals[NETWORK_EVENT], 0,
 		       event, connection);
 }
@@ -1338,9 +1375,9 @@ re_emit_tls_certificate_changed (SoupMessage    *msg,
 				 GParamSpec     *pspec,
 				 SoupConnection *conn)
 {
-        soup_message_set_tls_certificate (msg,
-                                          soup_connection_get_tls_certificate (conn),
-                                          soup_connection_get_tls_certificate_errors (conn));
+        soup_message_set_tls_peer_certificate (msg,
+                                               soup_connection_get_tls_certificate (conn),
+                                               soup_connection_get_tls_certificate_errors (conn));
 }
 
 void
@@ -1361,10 +1398,12 @@ soup_message_set_connection (SoupMessage    *msg,
 	if (!priv->connection)
 		return;
 
+        priv->last_connection_id = soup_connection_get_id (priv->connection);
+
 	g_object_add_weak_pointer (G_OBJECT (priv->connection), (gpointer*)&priv->connection);
-        soup_message_set_tls_certificate (msg,
-                                          soup_connection_get_tls_certificate (priv->connection),
-                                          soup_connection_get_tls_certificate_errors (priv->connection));
+        soup_message_set_tls_peer_certificate (msg,
+                                               soup_connection_get_tls_certificate (priv->connection),
+                                               soup_connection_get_tls_certificate_errors (priv->connection));
 
 	g_signal_connect_object (priv->connection, "event",
 				 G_CALLBACK (re_emit_connection_event),
@@ -1397,8 +1436,10 @@ soup_message_cleanup_response (SoupMessage *msg)
         soup_message_set_status (msg, SOUP_STATUS_NONE, NULL);
         soup_message_set_http_version (msg, priv->orig_http_version);
 
-        if (!priv->connection)
-                soup_message_set_tls_certificate (msg, NULL, 0);
+        if (!priv->connection) {
+                soup_message_set_tls_peer_certificate (msg, NULL, 0);
+                priv->last_connection_id = 0;
+        }
 
         g_object_thaw_notify (G_OBJECT (msg));
 }
@@ -1423,6 +1464,7 @@ soup_message_cleanup_response (SoupMessage *msg)
  *   and proxy authentication. Note that #SoupMessage::authenticate signal will
  *   be emitted, if you want to disable authentication for a message use
  *   soup_message_disable_feature() passing #SOUP_TYPE_AUTH_MANAGER instead.
+ * @SOUP_MESSAGE_COLLECT_METRICS: Metrics will be collected for this message.
  *
  * Various flags that can be set on a #SoupMessage to alter its
  * behavior.
@@ -1844,7 +1886,7 @@ soup_message_set_first_party (SoupMessage *msg,
         GUri *first_party_normalized;
 
 	g_return_if_fail (SOUP_IS_MESSAGE (msg));
-        g_return_if_fail (SOUP_URI_IS_VALID (first_party));
+        g_return_if_fail (first_party != NULL);
 
 	priv = soup_message_get_instance_private (msg);
         first_party_normalized = soup_uri_copy_with_normalized_flags (first_party);
@@ -1904,18 +1946,19 @@ soup_message_set_site_for_cookies (SoupMessage *msg,
 			           GUri        *site_for_cookies)
 {
 	SoupMessagePrivate *priv;
-        GUri *site_for_cookies_normalized;
+        GUri *site_for_cookies_normalized = NULL;
 
 	g_return_if_fail (SOUP_IS_MESSAGE (msg));
-        g_return_if_fail (SOUP_URI_IS_VALID (site_for_cookies));
 
 	priv = soup_message_get_instance_private (msg);
-        site_for_cookies_normalized = soup_uri_copy_with_normalized_flags (site_for_cookies);
-        if (!site_for_cookies_normalized)
-                return;
+        if (site_for_cookies) {
+                site_for_cookies_normalized = soup_uri_copy_with_normalized_flags (site_for_cookies);
+                if (!site_for_cookies_normalized)
+                        return;
+        }
 
 	if (priv->site_for_cookies) {
-		if (soup_uri_equal (priv->site_for_cookies, site_for_cookies_normalized)) {
+		if (site_for_cookies_normalized && soup_uri_equal (priv->site_for_cookies, site_for_cookies_normalized)) {
                         g_uri_unref (site_for_cookies_normalized);
                         return;
                 }
@@ -1973,18 +2016,18 @@ soup_message_get_is_top_level_navigation (SoupMessage *msg)
 }
 
 /**
- * soup_message_get_tls_certificate:
+ * soup_message_get_tls_peer_certificate:
  * @msg: a #SoupMessage
  *
- * Gets the #GTlsCertificate associated with @msg's connection.
+ * Gets the peer's #GTlsCertificate associated with @msg's connection.
  * Note that this is not set yet during the emission of
  * SoupMessage::accept-certificate signal.
  *
- * Returns: (transfer none) (nullable): @msg's TLS certificate,
+ * Returns: (transfer none) (nullable): @msg's TLS peer certificate,
  *    or %NULL if @msg's connection is not SSL.
  */
 GTlsCertificate *
-soup_message_get_tls_certificate (SoupMessage *msg)
+soup_message_get_tls_peer_certificate (SoupMessage *msg)
 {
 	SoupMessagePrivate *priv;
 
@@ -1992,21 +2035,21 @@ soup_message_get_tls_certificate (SoupMessage *msg)
 
 	priv = soup_message_get_instance_private (msg);
 
-	return priv->tls_certificate;
+	return priv->tls_peer_certificate;
 }
 
 /**
- * soup_message_get_tls_certificate_errors:
+ * soup_message_get_tls_peer_certificate_errors:
  * @msg: a #SoupMessage
  *
- * Gets the errors associated with validating @msg's TLS certificate.
+ * Gets the errors associated with validating @msg's TLS peer certificate.
  * Note that this is not set yet during the emission of
  * SoupMessage::accept-certificate signal.
  *
- * Returns: a #GTlsCertificateFlags with @msg's TLS certificate errors.
+ * Returns: a #GTlsCertificateFlags with @msg's TLS peer certificate errors.
  */
 GTlsCertificateFlags
-soup_message_get_tls_certificate_errors (SoupMessage *msg)
+soup_message_get_tls_peer_certificate_errors (SoupMessage *msg)
 {
         SoupMessagePrivate *priv;
 
@@ -2014,7 +2057,7 @@ soup_message_get_tls_certificate_errors (SoupMessage *msg)
 
 	priv = soup_message_get_instance_private (msg);
 
-	return priv->tls_certificate_errors;
+	return priv->tls_peer_certificate_errors;
 }
 
 /**
@@ -2309,3 +2352,96 @@ soup_message_set_is_options_ping (SoupMessage *msg,
         if (priv->is_options_ping)
                 soup_message_set_method (msg, SOUP_METHOD_OPTIONS);
 }
+
+/**
+ * soup_message_get_connection_id:
+ * @msg: The #SoupMessage
+ *
+ * Returns the unique idenfier for the last connection used.
+ * This may be 0 if it was a cached resource or it has not gotten
+ * a connection yet.
+ *
+ * Returns: An id or 0 if no connection.
+ */
+guint64
+soup_message_get_connection_id (SoupMessage *msg)
+{
+        SoupMessagePrivate *priv = soup_message_get_instance_private (msg);
+
+        g_return_val_if_fail (SOUP_IS_MESSAGE (msg), 0);
+
+        return priv->last_connection_id;
+}
+
+/**
+ * soup_message_get_metrics:
+ * @msg: The #SoupMessage
+ *
+ * Get the #SoupMessageMetrics of @msg. If the flag %SOUP_MESSAGE_COLLECT_METRICS is not
+ * enabled for @msg this will return %NULL.
+ *
+ * Returns: (transfer none) (nullable): a #SoupMessageMetrics, or %NULL
+ */
+SoupMessageMetrics *
+soup_message_get_metrics (SoupMessage *msg)
+{
+        SoupMessagePrivate *priv;
+
+        g_return_val_if_fail (SOUP_IS_MESSAGE (msg), NULL);
+
+        priv = soup_message_get_instance_private (msg);
+        if (priv->metrics)
+                return priv->metrics;
+
+        if (priv->msg_flags & SOUP_MESSAGE_COLLECT_METRICS)
+                priv->metrics = soup_message_metrics_new ();
+
+        return priv->metrics;
+}
+
+void
+soup_message_set_metrics_timestamp (SoupMessage           *msg,
+                                    SoupMessageMetricsType type)
+{
+        SoupMessageMetrics *metrics = soup_message_get_metrics (msg);
+        guint64 timestamp;
+
+        if (!metrics)
+                return;
+
+        timestamp = g_get_monotonic_time ();
+        switch (type) {
+        case SOUP_MESSAGE_METRICS_FETCH_START:
+                memset (metrics, 0, sizeof (SoupMessageMetrics));
+                metrics->fetch_start = timestamp;
+                break;
+        case SOUP_MESSAGE_METRICS_DNS_START:
+                metrics->dns_start = timestamp;
+                break;
+        case SOUP_MESSAGE_METRICS_DNS_END:
+                metrics->dns_end = timestamp;
+                break;
+        case SOUP_MESSAGE_METRICS_CONNECT_START:
+                metrics->connect_start = timestamp;
+                break;
+        case SOUP_MESSAGE_METRICS_CONNECT_END:
+                metrics->connect_end = timestamp;
+                break;
+        case SOUP_MESSAGE_METRICS_TLS_START:
+                metrics->tls_start = timestamp;
+                break;
+        case SOUP_MESSAGE_METRICS_REQUEST_START:
+                metrics->request_start = timestamp;
+                break;
+        case SOUP_MESSAGE_METRICS_RESPONSE_START:
+                /* In case of multiple requests due to a informational response
+                 * the response start is the first one.
+                 */
+                if (metrics->response_start == 0)
+                        metrics->response_start = timestamp;
+                break;
+        case SOUP_MESSAGE_METRICS_RESPONSE_END:
+                metrics->response_end = timestamp;
+                break;
+        }
+}
